--[[

                                                                                                       
  ,ad8888ba,                                            ad88888ba                                      
 d8"'    `"8b                                          d8"     "8b ,d                 ,d               
d8'                                                    Y8,         88                 88               
88            ,adPPYYba, 88,dPYba,,adPYba,   ,adPPYba, `Y8aaaaa, MM88MMM ,adPPYYba, MM88MMM ,adPPYba,  
88      88888 ""     `Y8 88P'   "88"    "8a a8P_____88   `"""""8b, 88    ""     `Y8   88   a8P_____88  
Y8,        88 ,adPPPPP88 88      88      88 8PP"""""""         `8b 88    ,adPPPPP88   88   8PP"""""""  
 Y8a.    .a88 88,    ,88 88      88      88 "8b,   ,aa Y8a     a8P 88,   88,    ,88   88,  "8b,   ,aa  
  `"Y88888P"  `"8bbdP"Y8 88      88      88  `"Ybbd8"'  "Y88888P"  "Y888 `"8bbdP"Y8   "Y888 `"Ybbd8"'  
                                                                                                       
                                                                                                       
]]

GameState = inherited("GameState", State)

function GameState:new(players)
	local o = instance(self)
	o.time = 0
	o.players = players
	o.playerKeyFade = 1
	o:createMap()
	o.spawnWaves = 0
	o.speedBonus = 0
	o.firstWave = false
	o.winTimer = 0
	if players then
		o:setPlayers(players)
	end
	o.offs = {{x = 0, y = 0, a = 0}, {x = 0, y = 0, a = 0}}
	return o
end

function GameState:init()
end

--[[
                                                                  
88        88                      88                               
88        88                      88              ,d               
88        88                      88              88               
88        88 8b,dPPYba,   ,adPPYb,88 ,adPPYYba, MM88MMM ,adPPYba,  
88        88 88P'    "8a a8"    `Y88 ""     `Y8   88   a8P_____88  
88        88 88       d8 8b       88 ,adPPPPP88   88   8PP"""""""  
Y8a.    .a8P 88b,   ,a8" "8a,   ,d88 88,    ,88   88,  "8b,   ,aa  
 `"Y8888Y"'  88`YbbdP"'   `"8bbdP"Y8 `"8bbdP"Y8   "Y888 `"Ybbd8"'  
             88                                                    
             88          

]]

function GameState:getPlayerPosition(i, w, h)
	if #self.players == 2 then
		return window.getSquareCorner(i, w, 0.5)
	elseif #self.players == 3 then
		local ang = (i-1)*math.pi*2/#self.players
		local less = math.min(window.w, window.h)*(1-math.min(w,h))*0.5
		return window.w*0.5 + math.cos(ang) * less, window.h*0.5 + math.sin(ang) * less
	else
		return window.getSquareCorner(i, w, h)
	end
end

function GameState:update(time)
	self.time = self.time + time

	if self.players then
		if self.playerKeyFade > 0 then
			self.playerKeyFade = math.max(0,self.playerKeyFade - time)
			if self.playerKeyFade <= 0 then
				self.waveTimer = 1
				self.extraWave = 0
			end
		end

		if self.map.hasOpponents then
			if self.waveTimer then
				self.waveTimer = self.waveTimer - time
				if self.waveTimer <= 0 then
					if not self.firstWave then
						self.firstWave = true
						audio.playSound("sfx/woh2.ogg")
					else
						audio.playSound("sfx/woh1.ogg", 0.5, 0, math.random()*0.2+0.7)
					end
					self.map.pulseTimer = 1
					self.map:addWave(Wave:new(nil, self.map.width*0.5, self.map.height*0.5, 95 + self.speedBonus/5, 1))
					self.spawnWaves = self.spawnWaves -1
					if self.spawnWaves > 0 then
						self.waveTimer = 1.75
					else
						self.waveTimer = nil
					end
				end
			end
		else
			self.map:removeWaves()
			self.winTimer = self.winTimer + time
			if #self.map.cells < 40 then
				for index, cell in pairs(self.map.cells) do
					if math.random() < time*0.1 then
						cell:split()
					end
				end
			end
		end
	end

	if math.random() < time*0.1 then
		audio.playSound("sfx/drop"..math.random(1,4)..".ogg",  math.random()*math.random()*0.5, math.random()*2-1,1)
	end

	if math.random() < time*0.03 then
		audio.playSound("sfx/ambience2.ogg")
	end

	if self.players then
		self.speedBonus = self.speedBonus + time
		if self.map then
			for index, player in pairs(self.players) do
				player.dead = true
				player.nextToDie = nil
			end
			self.map:update(time)
			for index, player in pairs(self.players) do
				if player.dead and not player.rejoinTimer then
					player.rejoinTimer = 20
				end
			end
			if self.extraWave and not self.map.fieldIsActive and self.spawnWaves <= 0 then
				self.spawnWaves = self.extraWave + 1
				self.extraWave = self.extraWave +1
				self.waveTimer = 1
			end
		end
		for index, player in pairs(self.players) do
			player:update(time)
			if player.rejoinTimer and self.map.hasOpponents then
				if player.rejoinTimer > 0 then
					local was = player.rejoinTimer
					player.rejoinTimer = player.rejoinTimer - time
					if math.floor(was) > math.floor(player.rejoinTimer) then
						local x,y = self:getPlayerPosition(index, 0.2, 0.15)
						audio.playSound("sfx/heartbeat2.ogg", math.max(0.2,1.2-player.rejoinTimer/10), (x-self.map.width*0.5)/self.map.width*2)
					end
					if player.rejoinTimer <= 0 then
						local x,y = self:getPlayerPosition(index, 0.2, 0.15)
						audio.playSound("sfx/heartbeat.ogg", 1.4, (x-self.map.width*0.5)/self.map.width*2)
						player.showRespawnText = 4
						player.respawnBonusTimer = (player.respawnBonusTimer or 0) + 1
					end
				end
			end
		end
	else
		self.map:update(time)
	end

	for index, offs in pairs(self.offs) do
		if math.random() < time*10 then
			offs.x = math.random()*2-1
			offs.y = math.random()*2-1
			offs.a = math.random()*5-5
		end
	end	
end

function GameState:setPlayers(players)
	states.get("music"):setNextMusic("ambience1.ogg", 0.2, true)
	_config.musicVolume = 0.2
	self.players = players
	for index, player in pairs(players) do
		player.rejoinTimer = nil
		player.nextToDie = nil
		player.respawnBonusTimer = nil
	end
	for index, player in pairs(self.players) do
		local x,y = self:getPlayerPosition(index, 0.2, 0.2)
		for i=1,5 do
			local cell = self.map:addCell(Cell:new(player, x + math.random(0,50)-25, y + math.random(0,50)-25, 17, 0, 0))
			cell:setNuclea(2)
			cell.startFade = math.random()*0.4-0.4
		end
	end
end

function GameState:createMap()
	self.map = Map:new(window.w, window.h)

	for i=1,math.random(1,2) do
		self.map:addLight(Light:new(self.map.width*math.random(), self.map.height*math.random()))	
	end

	for i=1,math.random(2,4) do
		self.map:addCreep(Creep:new("bigThing", math.random(0,self.map.width),math.random(0,self.map.height),0,0,math.randomAngle(),0.7+math.random()*0.6, 0.5, 20))	
	end

	for i=1,math.random(5,10) do
		self.map:addCreep(Creep:new("creep", math.random(0,self.map.width),math.random(0,self.map.height),0,0,math.randomAngle(),0.7+math.random()*0.6, 0.5, 40))	
	end
	if math.random() > 0.5 then
		self.map:addCreep(Creep:new("snake", math.random(0,self.map.width),math.random(0,self.map.height),0,0,math.randomAngle(),0.7+math.random()*0.6, 0.5, 80))
	end
	for i=1,math.random(30,50) do
		self.map:addCreep(Creep:new("particle", math.random(0,self.map.width),math.random(0,self.map.height),0,0,math.randomAngle(),0.7+math.random()*0.6, 0.5, 100))
	end
	self.map:sortCreeps()
end

--[[
                                                                     
88888888ba                                 88                        
88      "8b                                88                        
88      ,8P                                88                        
88aaaaaa8P' ,adPPYba, 8b,dPPYba,   ,adPPYb,88  ,adPPYba, 8b,dPPYba,  
88""""88'  a8P_____88 88P'   `"8a a8"    `Y88 a8P_____88 88P'   "Y8  
88    `8b  8PP""""""" 88       88 8b       88 8PP""""""" 88          
88     `8b "8b,   ,aa 88       88 "8a,   ,d88 "8b,   ,aa 88          
88      `8b `"Ybbd8"' 88       88  `"8bbdP"Y8  `"Ybbd8"' 88          
                                                                     

]]

function GameState:render()
	video.clearScreen(60,60,60)
	video.renderTiledSprite(SPRITES.bg, -2 + self.offs[1].x,-2 + self.offs[1].y, window.w+4, window.h+4, 1, 255,255,255,255)
	video.renderTiledSprite(SPRITES.bg, -22 + self.offs[2].x,-22 + self.offs[2].y, window.w+44, window.h+44, 1, 128 + self.offs[2].a,255,255,255)
	

	if self.map then
		self.map:bgRender()
		if self.players then
			for index, player in pairs(self.players) do
				if player.dead then
					local x,y = window.getCorner(index, 0, 0)
					video.renderAdditiveSpriteState(SPRITES.spotlight, x, y, 2, 0, 0.1*player.deadFade*player.color.r, 0.1*player.deadFade*player.color.g, 0.1*player.deadFade*player.color.b)
				end
			end
		end
		self.map:render()
	end

	video.renderFrameSprite(SPRITES.gradient, 0, 0, window.w, window.h, math.min(255,244 - math.cos(self.time)*9 - math.cos(self.time*6.211)*5 - math.sin(self.time*1.5353)*7), 255, 255, 255)
	
	if self.players then
		for index, player in pairs(self.players) do
			local fade = math.sCurve(self.playerKeyFade)
			local sx, sy = window.getCorner(index, 0.2, 0.15)
			local x,y = self:getPlayerPosition(index, 0.05, 0.05)
			x,y = x*(1-fade) + sx*fade,  y*(1-fade) + sy*fade
			player:renderInputKeyAt(x,y + 30*(fade), 0.75 + fade*0.25)
			if player.rejoinTimer then
				local x,y = self:getPlayerPosition(index, 0.2, 0.15)
				local scale =  math.max(1, (4-player.rejoinTimer/2)) + 0.5*math.max(0,math.cos(2*math.pi*player.rejoinTimer)*10-9)
				if player.rejoinTimer <= 0 then
					if player.showRespawnText and player.showRespawnText > 0 then
						local sc = 1+0.5*(player.respawnBonusTimer or 0)*(player.respawnBonusTimer or 0)
						video.renderTextSprites("PRESS TO SPAWN!", x,y+fonts.alignY("big",sc), 1, "big", 255-math.max(0,math.cos(self.time*4*math.pi)*150), player.color.r, player.color.g, player.color.b,nil,sc)
					end
				else
					video.renderTextSprites(math.ceil(player.rejoinTimer), x,y+fonts.alignY("big",scale ), 1, "big", 255, player.color.r, player.color.g, player.color.b,nil,scale)
				end
			end
			if player.wins and player.wins > 0 then
				local x,y = self:getPlayerPosition(index, 0.1, 0.1)
				if player.wins >= 2 then
					video.renderTextSprites("WINS "..player.wins, x, y + fonts.alignY("big", 1.5), 1, "big", 155 + math.min(100,math.max(0,100*math.cos(self.time*math.pi*2))*100), player.color.r, player.color.g, player.color.b, nil, 1.5)
				else
					video.renderTextSprites("WINS "..player.wins, x, y + fonts.alignY("medium"), 1, "medium", 255, player.color.r, player.color.g, player.color.b)
				end
			end
		end	

		if self.waveTimer and self.waveTimer > 0 then
			local radius = math.easeOut(self.waveTimer/2)*50
			video.renderSpriteCircleLine(self.map.width*0.5, self.map.height*0.5,radius, 1,1, 32, 0, math.max(0,math.min(255,(1-self.waveTimer)*255)), 0, 0, 0, math.min(radius,16), nil, SPRITES.wave)
		end



		if self.map.winner then
			local alpha = math.sCurve(math.min(1,self.winTimer))
			local x,y = window.getCorner(self.map.winner.id, 0.2 + 0.3*alpha,0.2 + 0.3*alpha)
			video.renderShadowedTextSprites("WINNER", x,y+fonts.alignY("big", 3)-40, 1, "big", alpha*255*math.abs(math.cos(self.winTimer*5)), self.map.winner.color.r, self.map.winner.color.g, self.map.winner.color.b, nil, 3)
			video.renderShadowedTextSprites("PLAYER "..self.map.winner.id, x,y+fonts.alignY("big", 2)+70, 1, "big", alpha*255*math.abs(math.cos(self.winTimer*5)), self.map.winner.color.r, self.map.winner.color.g, self.map.winner.color.b, nil, 2)
			if self.map.winner and self.map.winner.wins >= 3 then
				video.renderShadowedTextSprites("GRAND", x,y+fonts.alignY("big", 3)-240, 1, "big", alpha*255*math.abs(math.cos(self.winTimer*25)), self.map.winner.color.r, self.map.winner.color.g, self.map.winner.color.b, nil, 4)
			end
		end
	end
end
                                                                                                    
--[[
                                                          
88        88                         88                   
88        88                         88                   
88        88                         88                   
88aaaaaaaa88  ,adPPYba,   ,adPPYba,  88   ,d8  ,adPPYba,  
88""""""""88 a8"     "8a a8"     "8a 88 ,a8"   I8[    ""  
88        88 8b       d8 8b       d8 8888[      `"Y8ba,   
88        88 "8a,   ,a8" "8a,   ,a8" 88`"Yba,  aa    ]8I  
88        88  `"YbbdP"'   `"YbbdP"'  88   `Y8a `"YbbdP"'  
                                                          
                                                          
]]

function GameState:onPlayerAction(playerId, player)
	if self.winTimer >= 1 then
		if self.map.winner and self.map.winner.wins >= 3 then
			for index, player in pairs(self.players) do
				player.wins = 0
			end
		end
		states.destroy("game")
		states.add(GameState:new(self.players), "game")
	end
	player:action()
	
	if not self.map.winner and player.rejoinTimer then

		if player.rejoinTimer <= 0 then
			player.rejoinTimer = nil
			local x,y = self:getPlayerPosition(playerId, 0.2, 0.2)
			local nuclea = 2
			if player.respawnBonusTimer and player.respawnBonusTimer > 0 then
				local amt = math.ceil(player.respawnBonusTimer*player.respawnBonusTimer*3)
				audio.playSound("sfx/extra.ogg", 0.75 + amt/3, (x-self.map.width*0.5)/self.map.width*2, 0.8 + amt/6)
				nuclea = nuclea + amt
			end
			for i=1,nuclea do
				local cell = self.map:addCell(Cell:new(player, x + math.random(0,30)-15, y + math.random(0,30)-15, 17, 0, 0))
				cell:setNuclea(2)
			end
			return
		elseif player.rejoinTimer <= 1 then
			player.respawnBonusTimer = -5
		end
	end

	if self.map then
		local found = false
		for index, cell in pairs(self.map.cells) do
			if player == cell.owner then
				
				if cell.hitFactor > 0 then
					found = true
					local numOfCells = 0
					for index, ce in pairs(self.map.cells) do
						if ce.owner == player then
							numOfCells = numOfCells + 1
						end
					end
					local waves = self.map:getWavesAffecting(cell)
					for index, wave in pairs(waves) do
						wave:disableSegsAt(cell.x, cell.y, 225)
						if cell.hitFactor >= 0.75  - (cell.nuclea-1)*0.04 or cell.spiked then
							local angTo = math.angleBetweenPoints(cell.x, cell.y, wave.x, wave.y)
							local speed = 30
							if cell.spiked then
								speed = 120 + (cell.nuclea-5)*60
								audio.playSound("sfx/ring.ogg", 1, cell:getPan(), 0.6)
								cell.hitFeedback = 1
								if cell.spiked then
									cell.fadeIn = 0.9
								else
									cell.fadeIn = 0.8
								end
								cell.hitFactor = 0
								cell.lastHitFactor = 0
								cell:setNuclea(cell.nuclea +1)
								cell.radius = cell.radius +2
								cell.suckTimer = 2 + (cell.nuclea-1)*0.5
							end
							cell:addSpeed(math.cos(angTo)*speed, math.sin(angTo)*speed)
						end
					end
					if not cell.spiked then
						if cell.hitFactor >= 0.75 - (cell.nuclea-1)*0.04 then
							local speed = 90+ cell.nuclea*35 + numOfCells*5 + self.speedBonus/10 + math.max(0,(cell.lastWaveHit.speed-150))*0.5
							self.map:addWave(Wave:new(player, cell.x, cell.y, speed, cell.radius))
							audio.playSound("sfx/ring.ogg", 1, cell:getPan(), math.random()*0.1+0.4+(speed-90)/100)
							cell.hitFeedback = 1
							cell.fadeIn = 0.9
							cell.hitFactor = 0
							cell.lastHitFactor = 0
							cell:setNuclea(cell.nuclea +1)
							cell.radius = cell.radius +2
							cell.suckTimer = 2 + (cell.nuclea-1)*0.5
						else
							cell:split(1+self.speedBonus/120, 2+math.floor(cell.nuclea/3))
							audio.playSound("sfx/move"..math.random(1,18)..".ogg", 1, cell:getPan(), 1)
						end
					end
				end
				--if not found then
				--	if cell.waveProtected and cell.fadeIn >= 1 then
				--	found = true
				--	cell.fadeIn = 0.4
					--local angTo = math.angleBetweenPoints(cell.waveProtected.x, cell.waveProtected.y, cell.x, cell.y)
					--cell:addSpeed(math.cos(angTo)*4000/(cell.waveProtected.radius + 50), math.sin(angTo)*4000/(cell.waveProtected.radius + 50))
					--audio.playSound("sfx/move"..math.random(1,18)..".ogg", 0.4, cell:getPan(), 1.8)
					--cell.waveProtected:disableSegsAt(cell.x, cell.y, 100)
				--	end
				--end
			end
		end
		if not found then
			if player.nextToDie then
				player.nextToDie:suicide()
				for index, cell in pairs(self.map.cells) do
					local dist = (35 + math.distance(player.nextToDie.x, player.nextToDie.y, cell.x, cell.y))
					local factor = 1 + (player.nextToDie.nuclea-1)*0.5
					local ang = math.angleBetweenPoints(player.nextToDie.x, player.nextToDie.y, cell.x, cell.y)
					cell:addSpeed(math.cos(ang)*factor*1800/(dist), math.sin(ang)*factor*1800/(dist))
				end	
			end
		end
	end
end

function GameState:onInput(device, key, deviceId)
	if self.players then
		for index, player in pairs(self.players) do
			if player.input.device == device and player.input.key == key and player.input.deviceId == deviceId then
				self:onPlayerAction(index, player)
			end
		end
	end
end

function GameState:keyPressed(key)
	self:onInput("keyboard", key)
end

function GameState:mouseClick(x, y, button, clickCount)
	self:onInput("mouse", button)
end

function GameState:joyButtonPressed(joy, btn)
	self:onInput("joystick", btn, joy)
end