COLORS = {}
COLORS.white = {a = 255, r = 255, g = 255, b = 255}

Color = {}

Color.__ignore = {}
--Color.__ignoreMe = true

function Color.__tostring()
	return "Color"
end

function Color:new(a, r, g, b)

	local color = {}
	
	color.a = a
	color.r = r
	color.g = g
	color.b = b

	color = setmetatable(color, self)
	self.__index = self
	return color
end

Color.__newPackage = {
		vars = {
			{"a", 		"number"},
			{"r", 		"number"}, 
			{"g", 		"number"}, 
			{"b", 		"number"}
			},
		method = Color.new,
	}

function Color:newCopy(c)

	local color = {}
	
	color.a = c.a
	color.r = c.r
	color.g = c.g
	color.b = c.b

	color = setmetatable(color, self)
	self.__index = self
	return color
end

function Color:getColor(index)
	if index == 1 then
		return self.a
	elseif index == 2 then
		return self.r
	elseif index == 3 then
		return self.g
	else
		return self.b
	end
end

function Color:save(file)
	file:writeInt(self.a)
	file:writeInt(self.r)
	file:writeInt(self.g)
	file:writeInt(self.b)
end

function Color:load(file)

	local a,r,g,b = 0,0,0,0
	
	a = file:readInt()
	r = file:readInt()
	g = file:readInt()
	b = file:readInt()

	return Color:new(a,r,g,b)
end

FloatColor = {}

FloatColor.__ignore = {}
--FloatColor.__ignoreMe = true

function FloatColor.__tostring()
	return "FloatColor"
end

function FloatColor:new(a, r, g, b)

	local color = {}
	
	color.a = a
	color.r = r
	color.g = g
	color.b = b

	color = setmetatable(color, self)
	self.__index = self
	return color
end

FloatColor.__newPackage = {
		vars = {
			{"a", 		"number"},
			{"r", 		"number"}, 
			{"g", 		"number"}, 
			{"b", 		"number"}
			},
		method = FloatColor.new,
	}

function FloatColor:newCopy(c)

	local color = {}
	
	color.a = c.a
	color.r = c.r
	color.g = c.g
	color.b = c.b

	color = setmetatable(color, self)
	self.__index = self
	return color
end

function FloatColor:getColor(index)
	if index == 1 then
		return self.a
	elseif index == 2 then
		return self.r
	elseif index == 3 then
		return self.g
	else
		return self.b
	end
end

function FloatColor:save(file)
	file:writeFloat(self.a)
	file:writeFloat(self.r)
	file:writeFloat(self.g)
	file:writeFloat(self.b)
end

function FloatColor:load(file)

	local a,r,g,b = 0,0,0,0
	
	a = file:readFloat()
	r = file:readFloat()
	g = file:readFloat()
	b = file:readFloat()

	return FloatColor:new(a,r,g,b)
end



ColorQuad = {}

ColorQuad.__ignore = {}
--ColorQuad.__ignoreMe = true

function ColorQuad.__tostring()
	return "ColorQuad"
end

function ColorQuad:new(a, r, g, b)

	local color = {}
	
	if type(a) == "table" then
		color.a = a
	else
		color.a = {a, a, a, a}
	end
	if type(r) == "table" then
		color.r = r
	else
		color.r = {r, r, r, r}
	end
	if type(g) == "table" then
		color.g = g
	else
		color.g = {g, g, g, g}
	end
	if type(b) == "table" then
		color.b = b
	else
		color.b = {b, b, b, b}
	end
	
	color = setmetatable(color, self)
	self.__index = self
	return color
end

ColorQuad.__newPackage = {
		vars = {
			{"a", 		"table"},
			{"r", 		"table"}, 
			{"g", 		"table"}, 
			{"b", 		"table"}
			},
		method = ColorQuad.new,
	}

function ColorQuad:newCopy(c)

	local color = {}
	
	color.a = table.copy(c.a)
	color.r = table.copy(c.r)
	color.g = table.copy(c.g)
	color.b = table.copy(c.b)

	color = setmetatable(color, self)
	self.__index = self
	return color
end

function ColorQuad:getColorQuad(index)
	if index == 1 then
		return self.a
	elseif index == 2 then
		return self.r
	elseif index == 3 then
		return self.g
	else
		return self.b
	end
end

function ColorQuad:save(file)
	for i=1, 4 do
		file:writeInt(self.a[i])
		file:writeInt(self.r[i])
		file:writeInt(self.g[i])
		file:writeInt(self.b[i])
	end
end

function ColorQuad:load(file)

	local a,r,g,b = {},{},{},{}
	
	for i=1, 4 do
		a[i] = file:readInt()
		r[i] = file:readInt()
		g[i] = file:readInt()
		b[i] = file:readInt()
	end
	
	return ColorQuad:new(a,r,g,b)
end