window = 	{
			w = 0, h = 0, -- this is the current size of the app window
			ref_w = 0, ref_h = 0, -- this is the initial size of the app window, never changes
			} 
canvas = 	{
			w = 0, h = 0, -- w and h is the current viewable render size
			ref_w = 0, ref_h = 0 -- this is the reference screen size, never changes, independent of physical size but maintains aspect ratio
			}

function window.setSize(w,h)
	window.w = w
	window.h = h
	video.setScreenSize(window.w, window.h)
end

local corners = {
				{w = 0.2, h = 0.2},
				}

function window.getSquareCorner(id, w, h)

	local s = math.min(window.h, window.h)
	local wd = (window.w-s)*0.5
	local hd = (window.h-s)*0.5
	if id == 1 then
		return wd + s*(w or 0), hd+ s*(h or 0)
	elseif id == 2 then
		return wd + s - s*(w or 0), hd+ s*(h or 0)
	elseif id == 3 then
		return wd + s*(w or 0), hd+ s - s*(h or 0)
	elseif id == 4 then
		return wd + s - s*(w or 0), hd+ s - s*(h or 0)
	end
end


function window.getCorner(id, w, h)
	if id == 1 then
		return window.w*(w or 0), window.h*(h or 0)
	elseif id == 2 then
		return window.w - window.w*(w or 0), window.h*(h or 0)
	elseif id == 3 then
		return window.w*(w or 0), window.h - window.h*(h or 0)
	elseif id == 4 then
		return window.w - window.w*(w or 0), window.h - window.h*(h or 0)
	end
end

function canvas.setSize(w,h)
	canvas.w = w
	canvas.h = h
	video.setRenderScreenSize( canvas.w , canvas.h)
end

function canvas.setToRef()
	if canvas.w ~= canvas.ref_w or canvas.h ~= canvas.ref_h then 
		canvas.setSize(canvas.ref_w, canvas.ref_h)
	end
end

function canvas.setToWindow()
	if canvas.w ~= window.w or canvas.h ~= window.h then 
		canvas.setSize(window.w, window.h)
	end
end

function window.init_16_9_1280()
	-- select reference resolution based on aspect ratio
	if (canvas.ref_w * 9) == (canvas.ref_h * 17) then
		-- 17:9 (unusual)
		canvas.ref_w, canvas.ref_h = 1260, 664
	elseif (canvas.ref_w * 9) == (canvas.ref_h * 16) then
		-- HD 1080, HD 720
		canvas.ref_w, canvas.ref_h = 1280, 720
	elseif (canvas.ref_w * 3) == (canvas.ref_h * 5) then
		-- WXGA, WVGA
		canvas.ref_w, canvas.ref_h = 1280, 768
	elseif (canvas.ref_w * 10) == (canvas.ref_h * 16) then
		-- WQXGA, WUXGA, WSXGA, WXGA, CGA, 1440x900
		canvas.ref_w, canvas.ref_h = 1280, 800
	elseif (canvas.ref_w * 2) == (canvas.ref_h * 3) then
		-- NTSC, 1440x960 etc
		canvas.ref_w, canvas.ref_h = 1280, 854
	elseif (canvas.ref_w * 3) == (canvas.ref_h * 4) then
		-- Standard 4:3 ratio, 1024x768 and so on
		canvas.ref_w, canvas.ref_h = 1280, 960
	elseif (canvas.ref_w * 4) == (canvas.ref_h * 5) then
		-- SXGA, QSXGA
		canvas.ref_w, canvas.ref_h = 1280, 1024
	elseif canvas.ref_w == 1366 and canvas.ref_h == 768 then
		-- 1366x768 doesn't align correctly to 16:9, but *almost*
		canvas.ref_w, canvas.ref_h = 1280, 768
	else
		debug.pl(1, "Error: Unable to find supported aspect ratio: "..canvas.ref_w .. " x "..canvas.ref_h)
		canvas.ref_w, canvas.ref_h = 1280, 800
	end
	canvas.w, canvas.h = canvas.ref_w, canvas.ref_h
end

function window.init()
	window.ref_w, window.ref_h = video.getScreenSize()
	window.w, window.h  = window.ref_w, window.ref_h
	canvas.ref_w, canvas.ref_h = window.w, window.h
end