MenuState = inherited("MenuState", State)

function MenuState:new(players)
	local o = instance(self)
	o.time = 0
	o.players = players or {}
	o.startTimer = 30
	o.timerSize = 1
	return o
end

function MenuState:init()
	-- states.get("music"):setNextMusic("menu.ogg", 0, true)
	-- _config.musicVolume = 1
end


--[[
                                                                  
88        88                      88                               
88        88                      88              ,d               
88        88                      88              88               
88        88 8b,dPPYba,   ,adPPYb,88 ,adPPYYba, MM88MMM ,adPPYba,  
88        88 88P'    "8a a8"    `Y88 ""     `Y8   88   a8P_____88  
88        88 88       d8 8b       88 ,adPPPPP88   88   8PP"""""""  
Y8a.    .a8P 88b,   ,a8" "8a,   ,d88 88,    ,88   88,  "8b,   ,aa  
 `"Y8888Y"'  88`YbbdP"'   `"8bbdP"Y8 `"8bbdP"Y8   "Y888 `"Ybbd8"'  
             88                                                    
             88          

]]

function MenuState:update(time)
	self.time = self.time + time
	local was = self.startTimer
	if #self.players >= 2 then
		self.startTimer = math.max(0,self.startTimer - time)
	end
	if math.ceil(self.startTimer) < math.ceil(was) then
		self.timerSize = self.timerSize + 0.25
		audio.playSound("sfx/heartbeat1.ogg")
	end
	if self.startTimer <= 0 then
		states.destroy("menu")
		states.get("game"):setPlayers(self.players)
	end

	local timerSizeTarget = math.max(1, 1.5 - self.startTimer/5)
	if self.timerSize > timerSizeTarget then
		self.timerSize = math.max(timerSizeTarget, self.timerSize - time*self.timerSize - time*2)
	end
end

                                                                                                    
--[[
                                                          
88        88                         88                   
88        88                         88                   
88        88                         88                   
88aaaaaaaa88  ,adPPYba,   ,adPPYba,  88   ,d8  ,adPPYba,  
88""""""""88 a8"     "8a a8"     "8a 88 ,a8"   I8[    ""  
88        88 8b       d8 8b       d8 8888[      `"Y8ba,   
88        88 "8a,   ,a8" "8a,   ,a8" 88`"Yba,  aa    ]8I  
88        88  `"YbbdP"'   `"YbbdP"'  88   `Y8a `"YbbdP"'  
                                                          
                                                          
]]

function MenuState:inputPicked(device, key, deviceId)
	for index, player in pairs(self.players) do
		local input = player.input
		if input.device == device and input.key == key and input.deviceId == deviceId then
			self.startTimer = self.startTimer - 1
			self.timerSize = 2
			--player:randomizeColor()
			return
		end
	end
	if #self.players < 4 then
		local player = Player:new(#self.players + 1)
		player:setInput(device, key, deviceId)
		--player:randomizeColor()
		table.insert(self.players, player)
		if #self.players >= 4 then
			self.startTimer = math.min(self.startTimer, 3)
		end	
	end
end

function MenuState:keyPressed(key)
	self:inputPicked("keyboard", key)
end

function MenuState:mouseClick(x, y, button, clickCount)
	self:inputPicked("mouse", button)
end

function MenuState:joyButtonPressed(joy, btn)
	self:inputPicked("joystick", btn, joy)
end

function MenuState:render()
	video.renderTextSprites(math.max(0,math.ceil(self.startTimer)), window.w*0.5, window.h*0.7 + fonts.alignY("big", self.timerSize), 1, "big", 255,255,255,255, nil, self.timerSize)
	
	for p=1,4 do
		local x,y = window.getCorner(p, 0.2, 0.15)
		
		local player = self.players[p]
		if player then
			video.renderTextSprites("Player "..p, x,y  + fonts.alignY("medium"), 1, "medium", 255,player.color.r,player.color.g,player.color.b)
			video.renderTextSprites(Hints.getHint(player.input.device, player.input.key, player.input.deviceId), x,y  + fonts.alignY("medium") + 30, 1, "medium", 255,player.color.r,player.color.g,player.color.b)
		else
			video.renderTextSprites("Player "..p, x,y  + fonts.alignY("medium"), 1, "medium", 255,255,255,255)
			if p == #self.players +1 then
				video.renderTextSprites("PICK BUTTON!!!", x,y  + fonts.alignY("medium") + 30, 1, "medium", math.abs(math.cos(self.time*5))*255,255,255,255)
			else

			end
		end
	end	
end

function MenuState:close()
end

