local CLASSES = {}
local CLASS_METATABLES = {}
local CLASS_PARENTS = {}
local INSTANCE_METATABLES = {}
local WEAK_INSTANCE_METATABLES = {}

local function getWeakInstanceMetaTable(classTable)
	if not WEAK_INSTANCE_METATABLES[classTable] then
		WEAK_INSTANCE_METATABLES[classTable] = {__index=classTable, __mode = "v"}
	end
	return WEAK_INSTANCE_METATABLES[classTable]
end

function getClassName(classTable)
	return CLASSES[classTable]
end

function getClassNameOf(instance)
	local mt = getmetatable(instance)
	if mt then
		return CLASSES[mt.__index]
	end
end

function getClassTable(className)
	return CLASSES[className]
end

function getClassTableOf(instance)
	local mt = getmetatable(instance)
	if mt then
		return mt.__index
	end
end

function getInstanceMetaTable(classTable)
	return INSTANCE_METATABLES[classTable]
end

function class(className)
	if CLASSES[className] then
		print("Class "..className.. " already defined")
		return CLASSES[className]
	else
		local classTable = {}
		if not INSTANCE_METATABLES[classTable] then
			INSTANCE_METATABLES[classTable] = {__index=classTable}
		end

		CLASSES[className] = classTable
		CLASSES[classTable] = className
		return classTable
	end
end

function inherited(className, parentClassTable)
	if not parentClassTable then
		print(tostring(className).. " cannot inherit nil")
	end
	local classTable = class(className)
	if not CLASS_METATABLES[classTable] then
		CLASS_METATABLES[classTable] = {__index=parentClassTable}
		classTable.super = function()
			return parentClassTable
		end
	end
	return setmetatable(classTable, CLASS_METATABLES[classTable])
end

function weakInstance(classTable, ...)
	local a = setmetatable(classTable.super and weakInstance(classTable.super(),...) or {}, getWeakInstanceMetaTable(classTable))
	if classTable.onNew then
		classTable.onNew(a, ...)
	end
	debug.reportCreated(a)
	return a
end

function instance(classTable, ...)
	local a = setmetatable(classTable.super and instance(classTable.super(),...) or {}, INSTANCE_METATABLES[classTable])
	if classTable.onNew then
		classTable.onNew(a, ...)
	end
	debug.reportCreated(a)
	return a
end

function premade(classTable, baseTable, ...)
	local a = setmetatable(classTable.super and premade(classTable.super(), baseTable, ...) or baseTable or {}, INSTANCE_METATABLES[classTable])
	if classTable.onNew then
		classTable.onNew(a, ...)
	end
	debug.reportCreated(a)
	return a
end

State = {}

function State:onNew()
end

function State:update(time)

end

function State:init()

end

function State:close()

end

function State:enable()

end

function State:disable()

end

function State:hide()

end

function State:show()

end

function State:preRender()

end

function State:render()

end

function State:keyPressed(key)

end

function State:mouseClick(x,y,button,count)

end

function State:joyButtonPressed(joy, btn)

end

function State:joyButtonReleased(joy, btn)

end
